"""
create_container(container_config::Dict, container_name::String; auth_token::String="")

Creates and(or) starts a (existing) Docker container based on the provided configuration. If a container with the specified name already exists and is stopped, it will be started. If it doesn't exist, a new container will be created and the started.

# Arguments
- `container_config::Dict`: Configuration for the Docker container, including image, environment variables, and ports.
- `container_name::String`: The desired name for the container. Must be unique.
- `auth_token::String=""` (optional): Authentication token for Docker API access.

# Behavior
- Checks for an existing container with the given name.
- If the container exists and is stopped, it is started.
- If the container does not exist, a new one is created with the specified configuration.
- Prints a message indicating the action taken.

# Example
> container_config = Dict("Image" => "nginx:latest", "Env" => ["EXAMPLE_VAR=example_value"])
> container_name = "my_nginx_container"
> create_container(container_config, container_name)

This attempts to create or start a container named `my_nginx_container` using the `nginx:latest` image.
"""
function create_container(container_config::Dict, container_name::String; auth_token::String="")
    # List all containers (including stopped ones)
    all_containers = list_containers(; auth_token=auth_token)
    
    # Check if the container with the specified name exists
    existing_container = findfirst(c -> any(name -> name == "/$container_name", c["Names"]), all_containers)
    
    if existing_container !== nothing
        # Container exists, check if it's running
        container_id = all_containers[existing_container]["Id"]
        container_status = all_containers[existing_container]["State"]
        
        if container_status != "running"
            # Container is not running, start it
            println("Starting existing container: $container_name")
            docker_api_post("containers/$container_id/start", Dict(); auth_token=auth_token)
        else
            println("Container $container_name is already running.")
        end
    else
        # Container does not exist, create it
        println("Creating new container: $container_name")
        docker_api_post("containers/create?name=$container_name", container_config; auth_token=auth_token)
        all_containers = list_containers(; auth_token=auth_token)
        container_id = all_containers[findfirst(c -> any(name -> name == "/$container_name", c["Names"]), all_containers)]["Id"]
        docker_api_post("containers/$container_id/start", container_config; auth_token=auth_token)
    end
end

"""
remove_container(container_name::String; force::Bool=false, auth_token::String="")

Removes a Docker container specified by its name. Can forcefully remove the container if it is running.

# Arguments
- `container_name::String`: The name of the container to remove.
- `force::Bool=false`: Indicates whether to forcibly remove the container, including its volumes and any running instances.
- `auth_token::String=""` (optional): Authentication token for Docker API access.

# Behavior
- Checks for an existing container with the given name, including stopped containers.
- If the container exists, it attempts to remove it, forcibly if `force` is true.
- Prints a message indicating whether the container was removed or if it did not exist.

# Example
> container_name = "my_nginx_container"
> remove_container(container_name; force=true)

This removes the container named `my_nginx_container`, forcibly terminating it and removing its volumes if necessary.
"""
function remove_container(container_name::String; force::Bool=false, auth_token::String="")
    # List all containers (including stopped ones)
    all_containers = list_containers(; auth_token=auth_token)

    # Check if the container with the specified name exists
    existing_container = findfirst(c -> any(name -> name == "/$container_name", c["Names"]), all_containers)
    
    if existing_container !== nothing
        # Container exists, check if it's running
        container_id = all_containers[existing_container]["Id"]

        # Container is not running, start it
        println("Removing container: $container_name")
            # Endpoint to remove Docker containers
        endpoint = "containers/$container_id"

        # Parameters for the DELETE request
        params = Dict("force" => force, "v" => true)
        docker_api_post(endpoint, params; auth_token=auth_token)
    else
        # Container is not running, it can't be stopped
        println("Container $container_name does not exists.")
    end
end

"""
    list_containers(; all::Bool=true, auth_token::String="") -> Any

List all Docker containers on the host.

# Arguments
- `all::Bool=true`: Flag to list all containers. If `false`, only running containers are listed.
- `auth_token::String=""`: Optional authentication token.

# Returns
`Any`: A JSON array of container details as returned by the Docker API.

# Example
> containers = list_containers(all=true, auth_token="your_auth_token")

# Notes
- This function queries the 'containers/json' endpoint of the Docker API.
- The `docker_api_get` function is used to make the GET request with the specified parameters.
"""
function list_containers(; all::Bool=true, auth_token::String="")
    # Endpoint to list Docker containers
    endpoint = "containers/json"

    # Parameters for the GET request
    params = Dict("all" => all)

    # Call the docker_api_get function
    response = docker_api_get(endpoint, params; auth_token=auth_token)

    # The response should be a JSON array of containers
    return response  # or process further if needed
end

"""
stop_container(container_name::String; auth_token::String="")

Attempts to stop a running Docker container specified by its name. If the container is not found or is not running, a message is printed to indicate this.

# Arguments
- `container_name::String`: The name of the Docker container to stop.
- `auth_token::String=""` (optional): Authentication token for Docker API access.

# Behavior
- Searches for a running container with the specified name.
- If found and running, sends a request to stop the container.
- Prints a message indicating whether the container was stopped or if it was not running.

# Example
> stop_container("my_container")

Stops a running container named "my_container".
"""
function stop_container(container_name::String; auth_token::String="")
    # List all containers (including stopped ones)
    all_containers = list_containers(; all=false, auth_token=auth_token)

    # Check if the container with the specified name exists
    running_container = findfirst(c -> any(name -> name == "/$container_name", c["Names"]), all_containers)
    
    if running_container !== nothing
        # Container exists, check if it's running
        container_id = all_containers[running_container]["Id"]

        # Container is not running, start it
        println("Stopping existing container: $container_name")
        docker_api_post("containers/$container_id/stop", Dict(); auth_token=auth_token)
    else
        # Container is not running, it can't be stopped
        println("Container $container_name is not running.")
    end
end

"""
commit_container(container_name::String, repo_tag::String; auth_token::String="")

Commits the current state of a container to a new image with a specified repository tag. This is similar to taking a snapshot of the container's current state and saving it as an image.

# Arguments
- `container_name::String`: The name of the Docker container to commit as an image.
- `repo_tag::String`: The repository tag to apply to the new image. This should include the repository name and tag (e.g., "myrepo/myimage:latest").
- `auth_token::String=""` (optional): Authentication token for Docker API access.

# Behavior
- Searches for the specified container by name.
- If the container exists, commits its current state to a new image with the specified repository tag.
- Prints a message indicating the action taken or if the container does not exist.

# Example
> commit_container("my_container", "myrepo/myimage:latest")

Commits the state of "my_container" to a new image tagged as "myrepo/myimage:latest".
"""
function commit_container(container_name::String, repo::String, tag::String; auth_token::String="")
    # List all containers (including stopped ones)
    all_containers = list_containers(; auth_token=auth_token)

    # Check if the container with the specified name exists
    existing_container = findfirst(c -> any(name -> name == "/$container_name", c["Names"]), all_containers)
    
    if existing_container !== nothing
        # Container exists, check if it's running
        container_id = all_containers[existing_container]["Id"]

        # Container is not running, start it
        println("Comitting existing container: $container_name")
        docker_api_post("commit?container=$container_id&repo=$repo&tag=$tag", Dict(); auth_token=auth_token)
    else
        # Container is not running, it can't be stopped
        println("Container $container_name does not exists.")
    end
end
