"""
pull_image(image_name::String; auth_token::String="")

Pulls a Docker image from a registry. This function initiates a download of the specified image from its configured registry into the local Docker environment.

# Arguments
- `image_name::String`: The name of the Docker image to pull. This should include the tag if a specific version is desired (e.g., `ubuntu:latest`).
- `auth_token::String=""` (optional): An authentication token for accessing secured registries. The token is used to authenticate the pull request with the registry.

# Behavior
- Initiates a pull operation for the specified image. The function waits for the operation to complete, pulling the image into the local Docker environment.

# Example
> image_name = "nginx:latest"
> pull_image(image_name; auth_token="your_auth_token_here")

This example pulls the `nginx:latest` image from its registry, using the provided `auth_token` for authentication if necessary.
"""
function pull_image(image_name::String; auth_token::String="")
    # Endpoint to pull Docker images
    endpoint = "images/create"

    # Parameters for the POST request
    params = Dict("fromImage" => image_name)

    # Call the docker_api_post function with the auth_token
    docker_api_post(endpoint, params, auth_token=auth_token)
end

"""
remove_image(image_name::String; force::Bool=false, auth_token::String="")

Removes a Docker image from the local environment. This function can forcefully remove the image if it is in use or has dependent child images.

# Arguments
- `image_name::String`: The name (and optionally the tag) or ID of the Docker image to be removed.
- `force::Bool=false`: Whether to forcefully remove the image even if it is in use or has dependent child images.
- `auth_token::String=""` (optional): An authentication token. While typically not required for local operations, it might be necessary for certain secured environments.

# Behavior
- Attempts to remove the specified image from the local Docker environment. If `force` is set to `true`, it will remove the image even if there are complications such as the image being in use.

# Example
> image_name = "nginx:latest"
> remove_image(image_name; force=true, auth_token="your_auth_token_here")

This example forcefully removes the `nginx:latest` image from the local Docker environment, using the provided `auth_token` if necessary.
"""
function remove_image(image_name::String; force::Bool=false, auth_token::String="")
    # Endpoint to remove Docker images
    endpoint = "images/$image_name/remove"

    # Parameters for the DELETE request
    # 'force' parameter can be used to force removal of the image
    params = Dict("force" => force)

    # Call the docker_api_post function with the auth_token
    # Note: Depending on your Docker API, you might need to use a DELETE request instead
    docker_api_post(endpoint, params, auth_token=auth_token)
end

"""
    list_images(; auth_token::String="") -> Any

List all Docker images available in the local Docker image store.

# Arguments
- `auth_token::String=""`: An optional authentication token for accessing secured Docker resources.

# Returns
`Any`: A JSON array of images and their details as returned by the Docker API.

# Example
> images = list_images(auth_token="your_auth_token")

# Notes
- This function calls the Docker API's 'images/json' endpoint to retrieve a list of all images.
- The response is expected to be in JSON format and contains details about each image.
- The `docker_api_get` function is used to make the GET request. It should be properly implemented to handle GET requests and include authentication if provided.
"""
function list_images(; auth_token::String="")
    # Endpoint to list Docker images
    endpoint = "images/json"

    # Call the docker_api_get function
    response = docker_api_get(endpoint, Dict(); auth_token=auth_token)

    # The response should be a JSON array of images
    return response  # or process further if needed
end

"""
push_image(image_name::String; auth_token::String="")

Pushes a Docker image to a registry. This function requires the image to be already present in the local Docker environment and properly tagged to match the target registry's requirements.

# Arguments
- `image_name::String`: The name (and tag) of the Docker image to push to the registry. The name should include the registry path if pushing to a specific registry other than Docker Hub.
- `auth_token::String=""` (optional): An authentication token for accessing secured registries. The token is used to authenticate the push request with the registry.

# Behavior
- This function initiates a push operation for the specified image to its configured registry. The operation's success depends on network conditions, access permissions, and the existence of the image.

# Example
> image_name = "myregistry.com/myimage:latest"
> push_image(image_name; auth_token="my_auth_token")

This example pushes `myimage:latest` to `myregistry.com`, using `my_auth_token` for registry authentication.
"""
function push_image(image_name::String; auth_token::String="")
    # Endpoint to push the image to a registry
    push_endpoint = "images/$image_name/push"

    # Call the docker_api_post function to push the image
    docker_api_post(push_endpoint, Dict(), auth_token=auth_token)
end

"""
    tag_image(image_name::String, old_tag_name::String, new_image_name::String, new_tag_name::String; auth_token::String="")

Tags a Docker image with a new name and tag. This operation is local to the Docker environment and does not involve network communication with a registry.

# Arguments
- `image_name::String`: The name of the Docker image to be tagged. This does not include the tag of the source image.
- `old_tag_name::String`: The existing tag of the Docker image to be tagged. This specifies the exact version of the `image_name` to be used.
- `new_image_name::String`: The new name to apply to the image. This name can include a new registry path and repository but does not include the tag.
- `new_tag_name::String`: The new tag to apply to the image. This specifies the version of the `new_image_name`.
- `auth_token::String=""` (optional): While tagging typically does not require authentication, if this function is part of a broader workflow that includes registry interaction, an auth token can be specified.

# Behavior
- Applies the `new_image_name` and `new_tag_name` to the image identified by `image_name` and `old_tag_name`. If the operation is successful, the image will be available under both the old name/tag and the new name/tag.

# Example
> image_name = "myimage"
> old_tag_name = "oldtag"
> new_image_name = "myregistry.com/myimage"
> new_tag_name = "newtag"
> tag_image(image_name, old_tag_name, new_image_name, new_tag_name; auth_token="my_auth_token")

This example retags `myimage:oldtag` as `myregistry.com/myimage:newtag`, potentially preparing it for pushing to `myregistry.com`.
"""
function tag_image(image_name::String, old_tag_name::String, new_image_name::String, new_tag_name::String; auth_token::String="")
    # Endpoint to tag the image
    tag_endpoint = "images/$image_name:$old_tag_name/tag?repo=$new_image_name&tag=$new_tag_name"

    # Call the docker_api_post function to tag the image
    docker_api_post(tag_endpoint, Dict(), auth_token=auth_token)        
end