"""
    Module JuliaDockerClient

A Julia module for interacting with the Docker Engine API, enabling the management of Docker containers and images directly from Julia code.

This module provides a set of functions that allow users to perform various operations such as pulling and pushing images, creating, stopping, and removing containers, and more, using the Docker API.

## Dependencies
- HTTP.jl: For handling HTTP requests and responses.
- JSON.jl: For parsing and constructing JSON payloads.
- Sockets: For Unix socket communication with Docker.
- Base64: For encoding authentication tokens.

## Constants
- `DOCKER_API_MAIN_ENTRYPOINT`: The main entry point of the Docker API, defaulting to 'v1.40'.
- `DOCKER_API_HOST`: The host where the Docker API is accessible, defaulting to 'localhost'.
- `DOCKER_API_URL`: Constructs the full URL for the Docker API.
- `DOCKER_UNIX_SOCKET`: The path to the Unix socket for Docker, defaulting to '/run/user/&dollar;{UID}/podman/podman.sock'.

## Usage
The module provides various functions exported from included submodules. Before using these functions, ensure that the Docker Engine is running and accessible at the specified Unix socket or API URL.

### Exported Functions
- Docker API Interaction:
  - `test_connection`: Tests the connection to the Docker Unix socket.
  - `registry_login`: Generates an authentication token for Docker registry operations.
  - `prepare_container_resources_config`: Prepares the resource configuration for a container with constyrained resources (CPUs and memory in GB).
- Image Management:
  - `pull_image`: Pulls a Docker image from a registry.
  - `list_images`: Lists Docker images.
  - `remove_image`: Removes a Docker image.
  - `push_image`: Pushes a Docker image to a registry.
- Container Management:
  - `create_container`: Creates a new Docker container.
  - `stop_container`: Stops a running Docker container.
  - `list_containers`: Lists all Docker containers.
  - `remove_container`: Removes a Docker container.
  - `commit_container`: Commits changes in a container to create a new image.

## Configuration
The module can be configured via environment variables to change the Docker API endpoint and the Unix socket path. If these environment variables are not set, defaults are used.

## Notes
- Ensure that the Docker Engine is accessible and that appropriate permissions are set for the Unix socket or API endpoint.
- The module is intended for use with standard Docker installations, but it can also interact with other container management systems like Podman, provided they are compatible with the Docker API.
"""
module JuliaDockerClient

    using HTTP, JSON, Sockets, Base64

    const DOCKER_API_MAIN_ENTRYPOINT = get(ENV, "DOCKER_API_MAIN_ENTRYPOINT", "v1.40")

    const DOCKER_API_HOST = get(ENV, "DOCKER_API_HOST", "localhost")

    """
        DOCKER_API_URL

    Default Docker API URL. Can be overridden by setting the DOCKER_API_MAIN_ENTRYPOINT and DOCKER_API_HOST environment variables.
    """
    const DOCKER_API_URL = "http://$DOCKER_API_HOST/$DOCKER_API_MAIN_ENTRYPOINT/"

    const DOCKER_UNIX_SOCKET = get(ENV, "DOCKER_UNIX_SOCKET", haskey(ENV, "XDG_RUNTIME_DIR") ? joinpath(ENV["XDG_RUNTIME_DIR"], "podman/podman.sock") : "default_fallback_path")


    include("Utils.jl")
    export test_connection, registry_login, prepare_container_resources_config
    include("Images.jl")
    export pull_image, list_images, remove_image, push_image, tag_image
    include("Containers.jl")
    export create_container, stop_container, list_containers, remove_container, commit_container
    

end
