"""
    unix_socket_get(http_request::String, auth_token::String) -> String

Send a GET request through a Unix socket to the Docker API.

# Arguments
- `http_request::String`: The specific API endpoint and query string for the GET request.
- `auth_token::String`: Authentication token for registry access. If the token is empty, no authentication header is added.

# Returns
`String`: The raw HTTP response as a string.

# Usage
```julia
response = unix_socket_get("images/json", "auth_token_here")
```
This function connects to the Docker Unix socket, constructs a GET HTTP request with the specified endpoint and optional authentication, sends the request, and returns the raw response as a string. The response includes both the HTTP headers and the body.

# Notes
- The DOCKER_UNIX_SOCKET and DOCKER_API_MAIN_ENTRYPOINT are assumed to be globally defined constants indicating the path to the Unix socket and the main entry point of the Docker API, respectively.
- This function does not parse the response; it returns the raw HTTP response for further processing.
"""
function unix_socket_get(http_request::String, auth_token::String)
    sock = Sockets.connect("$DOCKER_UNIX_SOCKET")
    auth_header = isempty(auth_token) ? "" : "X-Registry-Auth: $auth_token\r\n"
    long_http_request = "GET /$DOCKER_API_MAIN_ENTRYPOINT/$http_request HTTP/1.1\r\nHost: $DOCKER_API_HOST\r\n$auth_header\r\n"
    response = ""  # Initialize response
    try
        write(sock, long_http_request)
        response = read(sock, String)
    finally
        close(sock)
    end
    return response
end

"""
    unix_socket_post(http_request::String, post_data::String, content_type::String="application/json", auth_token::String) -> String

Send a POST request through a Unix socket to the Docker API.

# Arguments
- `http_request::String`: The specific API endpoint for the POST request.
- `post_data::String`: The data to be sent in the body of the POST request.
- `content_type::String="application/json"`: The content type of the POST request. Defaults to "application/json".
- `auth_token::String`: Authentication token for registry access. It is included in the request's headers.

# Returns
`String`: The raw HTTP response as a string.

# Example
> response = unix_socket_post("containers/create", "{\"Image\": \"alpine\", \"Cmd\": [\"echo\", \"hello world\"]}", "application/json", "auth_token_here")

This function connects to the Docker Unix socket and constructs a POST HTTP request with the specified endpoint, data, and optional authentication. It sends the request and returns the raw response as a string. The response includes both the HTTP headers and the body.

# Notes
- The `DOCKER_UNIX_SOCKET` and `DOCKER_API_MAIN_ENTRYPOINT` are assumed to be globally defined constants indicating the path to the Unix socket and the main entry point of the Docker API, respectively.
- The function checks for a "200 OK" status in the response. If not found, it raises an error with the response details.
- This function does not parse the response; it returns the raw HTTP response for further processing.
"""
function unix_socket_post(http_request::String, post_data::String, auth_token::String; content_type::String="application/json")
    sock = Sockets.connect("$DOCKER_UNIX_SOCKET")

    # If post_data is supposed to be an empty JSON object but isn't provided as such, correct it
    if post_data == "" || post_data == "{}"
        post_data = "{}"
        post_data_length = length(post_data)
    else
        post_data_length = length(post_data)
    end
    
    # Include the auth_token in the header if it's provided
    auth_header = isempty(auth_token) ? "" : "X-Registry-Auth: $auth_token\r\n"
    
    long_http_request = "POST /$DOCKER_API_MAIN_ENTRYPOINT/$http_request HTTP/1.1\r\nHost: $DOCKER_API_HOST\r\n$auth_header Content-Type: $content_type\r\nContent-Length: $post_data_length\r\n\r\n$post_data"
    #println("Long request: $long_http_request")
    response = ""  # Initialize response
    try
        write(sock, long_http_request)
        response = read(sock, String)
    finally
        close(sock)
    end

    # Extract the status code from the response
    status_code_match = match(r"HTTP/1\.1 (\d{3})", response)
    if status_code_match !== nothing
        status_code = parse(Int, status_code_match.captures[1])

        # Check if the status code indicates success
        if 200 <= status_code < 300
            return response
        else
            error("Request failed with response: $response")
        end
    else
        error("Invalid HTTP response format: $response")
    end
end

"""
    docker_api_get(endpoint::String, params::Dict=Dict(); auth_token::String="") -> Any

Send a GET request to the Docker API via a Unix socket, particularly for querying Docker resources.

# Arguments
- `endpoint::String`: The Docker API endpoint to which the GET request is sent.
- `params::Dict=Dict()`: A dictionary of query parameters to append to the endpoint. Defaults to an empty dictionary.
- `auth_token::String=""`: An optional authentication token used for accessing secured Docker resources.

# Returns
`Any`: The response body parsed from JSON format. The actual return type depends on the specific endpoint and the structure of the response.

# Example
> images = docker_api_get("images/json", Dict("all" => "true"), auth_token="your_auth_token")

This function constructs the full URL for the request, including query parameters, and calls `unix_socket_get` to send the request through a Unix socket. It then extracts the body of the response and parses it from JSON.

# Notes
- Ensure that `unix_socket_get` and `get_response_body` are properly implemented and accessible.
- The function assumes that the response body is in JSON format and parses it accordingly. Non-JSON responses will lead to parsing errors.
- The `HTTP.escapeuri` function is used to correctly encode the query parameters.
- Authentication is managed via the `auth_token`, which is incorporated into the request header if provided.
"""
function docker_api_get(endpoint::String, params::Dict=Dict(); auth_token::String="")
    path = "$endpoint"
    if !isempty(params)
        path *= "?" * HTTP.escapeuri(params)
    end
    response = unix_socket_get(path, auth_token)
    response_body = get_response_body(response)
    return JSON.parse(response_body)
end

"""
    docker_api_post(endpoint::String, params::Dict=Dict(); auth_token::String="") -> Any

Send a POST request to the Docker API via a Unix socket, commonly used for actions like creating or modifying Docker resources.

# Arguments
- `endpoint::String`: The Docker API endpoint to which the POST request is sent.
- `params::Dict=Dict()`: A dictionary of data to be sent as the body of the POST request. Defaults to an empty dictionary.
- `auth_token::String=""`: An optional authentication token for secured Docker operations.

# Returns
`Any`: The parsed response body from JSON format. The actual return type depends on the specific endpoint and the response structure.

# Example
> response = docker_api_post("containers/create", Dict("Image" => "alpine", "Cmd" => ["echo", "hello world"]), auth_token="your_auth_token")

This function serializes the provided parameters into a JSON string for the POST request body. It then calls `unix_socket_post` to send the request and handles the response, extracting and parsing the response body from JSON.

# Notes
- `unix_socket_post` is used to send the request and must be implemented to handle POST requests correctly.
- The function assumes the response body is in JSON format. Non-JSON responses will cause parsing errors.
- It checks for a "200 OK" status within `unix_socket_post` and will raise an error for non-200 responses.
- Authentication is managed via the `auth_token`, which is included in the request if provided.
"""
function docker_api_post(endpoint::String, params::Dict=Dict(); auth_token::String="")
    # Serialize the parameters to JSON for the POST request body
    post_data = isempty(params) ? "" : JSON.json(params)

    # Send the POST request using unix_socket_post
    response = unix_socket_post(endpoint, post_data, auth_token)

    # Extract the status code from the response
    status_code_match = match(r"HTTP/1\.1 (\d{3})", response)
    status_code = status_code_match !== nothing ? parse(Int, status_code_match.captures[1]) : 0

    # The response is already checked for a 200 OK status in unix_socket_post
    response_body = get_response_body(response)
     
    # Check if the response body is empty and if the status code indicates success
    if isempty(response_body) && (200 <= status_code < 300)
        # For successful operations with no response body, return an empty Dict or a success message
        return Dict("success" => true)
    else
        # Attempt to parse each JSON object in the response body separately
        # Split the response body into individual JSON objects
        json_objects = split(response_body, "}\n{")
        parsed_responses = []
        
        for json_object in json_objects
            # Ensure proper JSON format after splitting
            json_object = strip(json_object)
            if !startswith(json_object, "{")
                json_object = "{" * json_object
            end
            if !endswith(json_object, "}")
                json_object = json_object * "}"
            end
            
            try
                # Parse each JSON object and add to the list
                push!(parsed_responses, JSON.parse(json_object))
            catch e
                println("Error parsing JSON object: $e")
                # Handle parsing error or log it
            end
        end
        
        # Return the list of parsed JSON objects
        return parsed_responses
    end
end


"""
    get_response_body(response::String) -> String

Parse the body from a chunked HTTP response string.

This function handles chunked transfer encoding of HTTP responses. It splits the response into chunks based on the chunk size specified in the response and reconstructs the full body.

# Arguments
- `response::String`: The raw HTTP response string, including headers and the chunked body.

# Returns
- `String`: The parsed response body.

# Example
> response = "HTTP/1.1 200 OK\r\nContent-Type: text/plain\r\nTransfer-Encoding: chunked\r\n\r\n5\r\nHello\r\n0\r\n\r\n"
> body = get_response_body(response)
> println(body)  # Output: Hello

In this example, `response` is a simulated raw HTTP response with a chunked body. The `get_response_body` function parses the response and extracts the body content.
"""
function get_response_body(response::String)
    parts = split(response, "\r\n\r\n", limit=2)
    if length(parts) < 2
        error("Invalid HTTP response format")
    end

    chunked_body = parts[2]
    body = ""

    while !isempty(chunked_body)
        chunk_parts = split(chunked_body, "\r\n", limit=2)
        if length(chunk_parts) < 2
            break
        end

        # Trim and check the chunk size part
        chunk_size_str = strip(chunk_parts[1])
        if isempty(chunk_size_str)
            break
        end

        chunk_size = try
            parse(Int, chunk_size_str, base=16)
        catch
            error("Failed to parse chunk size: '$chunk_size_str'")
        end

        if chunk_size == 0
            break
        end

        if length(chunk_parts[2]) < chunk_size
            error("Chunk size larger than remaining data")
        end

        chunk_content = chunk_parts[2][1:chunk_size]
        body *= chunk_content

        chunked_body = chunk_parts[2][chunk_size + 2:end] # +2 to skip the "\r\n" after the chunk
    end

    return body
end



"""
    registry_login(registry_url::String="registry.git.nilu.no", username::String="FACT_token", password::String="glpat-1zG-TQx___xLsPjj2vsG") -> String

Create a base64-encoded authentication token for Docker registry operations.

This function generates an authentication token that can be used in the `X-Registry-Auth` header for Docker API requests that require authentication, such as pushing or pulling private images.

# Arguments
- `registry_url::String="registry.git.nilu.no"`: The URL of the Docker registry. Defaults to "registry.git.nilu.no".
- `username::String="FACT_token"`: The username for the Docker registry.
- `password::String="glpat-1zG-TQx___xLsPjj2vsG"`: The password or token for the Docker registry.

# Returns
`String`: A base64-encoded JSON string that contains the authentication details.

# Example
> encoded_auth = registry_login("registry.example.com", "my_username", "my_password")

# Notes
- The function constructs a dictionary with the provided credentials and the registry URL, then converts this dictionary into a JSON string.
- The JSON string is encoded in base64, which is the format expected by the Docker Engine API for authentication headers.
- Be cautious with hardcoding credentials. In a production environment, consider secure ways to handle sensitive data like passwords or tokens.
"""
function registry_login(registry_url::String="registry.git.nilu.no", username::String="FACT_token", password::String="glpat-1zG-TQx___xLsPjj2vsG")
    # Create the auth details as a dictionary
    auth_details = Dict(
        "username" => username,
        "password" => password,
        "email" => "",  # Assuming email is not required; adjust as needed
        "serveraddress" => registry_url
    )

    # Convert the dictionary to a JSON string
    json_auth = JSON.json(auth_details)

    # Encode the JSON string in base64
    encoded_auth = base64encode(json_auth)

    # Now you can use encoded_auth as the value for the X-Registry-Auth header
    # in your HTTP request to the Docker Engine API.
    # Example: Using this in a POST request to /images/(name)/push

    return encoded_auth
end

"""
    test_connection() -> Bool

Test the connection to the Docker Unix socket.

This function attempts to establish a connection to the Docker Unix socket, indicated by the `DOCKER_UNIX_SOCKET` global constant. It is used to verify if the Docker daemon is accessible and the Unix socket is available.

# Returns
`Bool`: `true` if the connection is successfully established, `false` otherwise.

# Example
> is_connected = test_connection()

# Notes
- The function tries to open a socket connection using the path specified in `DOCKER_UNIX_SOCKET`.
- If the connection is successful, the socket is closed immediately, and the function returns `true`.
- If any error occurs during the connection attempt (such as the socket not being available), the function catches the exception, prints an error message, and returns `false`.
- This function is useful for preliminary checks to ensure that the Docker environment is set up correctly and the API is accessible.
"""
function test_connection() 
    try
        sock = Sockets.connect("$DOCKER_UNIX_SOCKET")
        close(sock)
        return true
    catch e
        println("Failed to connect to Unix socket: $e")
        return false
    end
end

"""
    parse_and_print_status_messages(json_response::String)

Parse and print status messages from a JSON-encoded response string.

This function is designed to process a JSON string, typically returned from Docker API calls, and print out status messages. It is particularly useful for handling streaming responses from the Docker API, such as the output from image pull or build operations.

# Arguments
- `json_response::String`: A string containing a JSON-encoded response from the Docker API.

# Usage
> parse_and_print_status_messages(response_json)

# Notes
- The function expects `json_response` to be a valid JSON string.
- It assumes that the JSON response is either an array of objects, each with a "status" field, or a single JSON object.
- If the response is an array, it iterates through each element, printing the value of the "status" field.
- If the response does not conform to the expected format (e.g., it is not an array), the function prints an "Unexpected response format" message.
- This function is useful for scenarios where real-time feedback from a Docker operation is required.
"""
function parse_and_print_status_messages(json_response::String)
    # Parse the JSON response string
    parsed_response = JSON.parse(json_response)

    # Check if the parsed response is an array (as expected for a stream of updates)
    if typeof(parsed_response) == Array
        for update in parsed_response
            # Each update is expected to be a JSON object with a "status" field
            status_message = get(update, "status", "No status message")
            println(status_message)
        end
    else
        println("Unexpected response format")
    end
end

"""
    prepare_container_resources_config(cores::Int, memory_gb::Int) -> Dict

Generate a container configuration dictionary with specified CPU and memory resource limits.

This function prepares a configuration dictionary for a Docker container with specific CPU core and memory settings. The memory limit is specified in gigabytes (GB) and internally converted to bytes, as Docker requires memory limits to be specified in bytes.

# Arguments
- `cores::Int`: The number of CPU cores to allocate to the container.
- `memory_gb::Int`: The memory limit for the container in gigabytes.

# Returns
- `Dict`: A container configuration dictionary with `HostConfig` settings for CPU and memory resources.

# Example
> cores = 4
> memory_gb = 2
> container_config = prepare_container_resources_config(cores, memory_gb)

This example creates a container configuration for a Docker container with 4 CPU cores and 2 GB of memory.
"""
function prepare_container_resources_config(cores::Int, memory_gb::Int)
    # Convert memory from GB to bytes (1 GB = 2^30 bytes)
    memory_bytes = memory_gb * 2^30

    # Prepare the HostConfig section of the container configuration
    host_config = Dict(
        "CpuCount" => cores,            # Set the number of cores
        "Memory"   => memory_bytes      # Set the memory limit
    )

    # Prepare the full container configuration
    container_config = Dict(
        "HostConfig" => host_config
        # Add other necessary configuration options here
    )

    return container_config
end
